/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <locale.h>
#include <math.h>
#include "data.h"
#include "$(appName).h"
#include "display.h"

#define STR_DISP_NUMBER_IS_TOO_LONG "Result overflow"
#define STR_DISP_INTERNAL_ERROR "INTERNAL ERROR!"
#define STR_DISP_DIVISION_BY_ZERO "Division by zero"
#define STR_DISP_UNDEFINED_VALUE "Undefined value"

#define DEFAULT_DEC_POINT_CHAR '.'
#define DEC_POINT_STR_SIZE 2

#define CALC_ZERO_EPS 1.0e-24

static struct data_info {
	bool error;
	result_type last_error;
	bool dirty;
	bool num_is_new;
	bool in_edit_mode;
	key_type op;
	double num;
	double acc;
	double rep;
	char dec_point_str[DEC_POINT_STR_SIZE];
} s_info = {
	.error = false,
	.last_error = RES_OK,
	.dirty = false,
	.num_is_new = false,
	.in_edit_mode = false,
	.op = KEY_ADD,
	.num = 0.0,
	.acc = 0.0,
	.rep = 0.0,
	.dec_point_str = {DEFAULT_DEC_POINT_CHAR,},
};

static void _reset(void);
static bool _set_special_key(key_type key);
static bool _set_numeric_key(key_type key);
static bool _set_operation_key(key_type key);
static bool _set_result_key(void);
static bool _set_sign_key(void);
static bool _set_erase_key(void);
static void _set_display_error(void);
static void _unset_edit_mode(void);
static bool _execute_operation(double rv);
static void _negate_value(void);

/**
 * @brief Initializes the calculator data structure.
 */
void data_initialize(void)
{
	s_info.last_error = RES_OK;

	_reset();
	data_update_region_format();
}

/**
 * @brief Registers the callback function to be invoked on display string change.
 * @param[in] cb The callback function pointer (NULL to unregister callback).
 */
void data_set_display_change_cb(display_changed_cb cb)
{
	display_set_changed_cb(cb);
}

/**
 * @brief Handles calculator key press.
 * @param[in] key Key value to handle.
 * @return This function returns 'true' if the key is set successfully,
 * otherwise 'false' is returned.
 */
bool data_set_key_press(key_type key)
{
	s_info.last_error = RES_OK;

	if (key == KEY_RESET) {
		_reset();
		return true;
	}

	if (s_info.error) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Calculator is in error state");
		s_info.last_error = RES_ILLEGAL_STATE;
		return false;
	}

	switch (key & KEY_TYPE_MASK) {
	case KEY_TYPE_NUM:
		_set_numeric_key(key);
		break;
	case KEY_TYPE_OP:
		_set_operation_key(key);
		break;
	default:
		_set_special_key(key);
		break;
	}

	if (s_info.last_error != RES_OK) {
		_set_display_error();
		return false;
	}

	return true;
}

/**
 * @brief Gets current display string message.
 * @return Pointer to the display string.
 */
const char *data_get_display_str(void)
{
	s_info.last_error = RES_OK;

	return display_get_str();
}

/**
 * @brief Updates regional formatting settings (decimal point character).
 */
void data_update_region_format(void)
{
	struct lconv *lc = localeconv();
	char new_dec_point_char;

	s_info.last_error = RES_OK;

	new_dec_point_char = (lc && lc->decimal_point && ('\0' != lc->decimal_point[0])) ?
							lc->decimal_point[0] : DEFAULT_DEC_POINT_CHAR;

	display_set_dec_point_char(new_dec_point_char);
	s_info.dec_point_str[0] = new_dec_point_char;
}

/**
 * @brief Gets current decimal point character depending on current regional settings.
 * @return Pointer to the string containing decimal point.
 */
const char *data_get_dec_point_str(void)
{
	s_info.last_error = RES_OK;

	return s_info.dec_point_str;
}

/**
 * @brief Function returns the value of recently occurred error.
 * @return Last error value.
 */
result_type data_get_last_error(void)
{
	return s_info.last_error;
}

/**
 * @brief Internal function which initializes the private data structure.
 * The data is reset to the initial state.
 */
static void _reset(void)
{
	s_info.error = false;
	s_info.dirty = false;
	s_info.num_is_new = false;
	s_info.in_edit_mode = false;
	s_info.op = KEY_ADD;
	s_info.num = 0.0;
	s_info.acc = 0.0;
	s_info.rep = 0.0;

	display_reset();
}

/**
 * @brief Internal function which sets one of the following keypad keys:
 * result key, sign key and erase key.
 * @param[in] key The key type to set.
 * @return This function returns 'true' if the key is set successfully,
 * otherwise 'false' is returned.
 */
static bool _set_special_key(key_type key)
{
	if (key == KEY_RESULT && !_set_result_key())
		return false;
	else if (key == KEY_SIGN && !_set_sign_key())
		return false;
	else if (key == KEY_ERASE && !_set_erase_key())
		return false;
	else
		s_info.last_error = RES_ILLEGAL_ARGUMENT;

	return (s_info.last_error == RES_OK);
}

/**
 * @brief Internal function which sets one of the numerical keypad keys (0-9).
 * @param[in] key The key type to set.
 * @return This function returns 'true' if the key is set successfully,
 * otherwise 'false' is returned.
 */
static bool _set_numeric_key(key_type key)
{
	bool reset = false;

	if (!s_info.in_edit_mode) {
		s_info.in_edit_mode = true;
		reset = true;
	}

	s_info.last_error = display_enter_key(key, reset);

	return (s_info.last_error == RES_OK);
}

/**
 * @brief Internal function which sets one of the arithmetical operation
 * keypad keys (addition, subtraction, multiplication, division).
 * @param[in] key The key type to set.
 * @return This function returns 'true' if the key is set successfully,
 * otherwise 'false' is returned.
 */
bool _set_operation_key(key_type key)
{
	bool res = true;

	if (key != KEY_ADD && key != KEY_SUB &&
		key != KEY_MUL && key != KEY_DIV) {
		s_info.last_error = RES_ILLEGAL_ARGUMENT;
		return false;
	}

	_unset_edit_mode();

	if (!s_info.dirty) {
		s_info.dirty = true;
		s_info.acc = s_info.num;
	} else if (s_info.num_is_new) {
		res = _execute_operation(s_info.num);
	}

	s_info.num_is_new = false;
	s_info.op = key;

	return res;
}

/**
 * @brief Internal function which sets the result (=) key.
 * @return This function returns 'true' if the key is set successfully,
 * otherwise 'false' is returned.
 */
static bool _set_result_key(void)
{
	_unset_edit_mode();

	if (s_info.dirty) {
		s_info.dirty = false;
		s_info.rep = s_info.num;
	} else if (s_info.num_is_new) {
		s_info.acc = s_info.num;
	}

	s_info.num_is_new = false;

	return _execute_operation(s_info.rep);
}

/**
 * @brief Internal function which sets the sign (+/-) key.
 * @return This function returns 'true' if the key is set successfully,
 * otherwise 'false' is returned.
 */
static bool _set_sign_key(void)
{
	if (s_info.in_edit_mode)
		display_negate();
	else
		_negate_value();

	return true;
}

/**
 * @brief Internal function which sets the erase (<-) key.
 * @return This function returns 'true' if the key is set successfully,
 * otherwise 'false' is returned.
 */
static bool _set_erase_key(void)
{
	if (s_info.in_edit_mode) {
		s_info.last_error = display_erase();
	}

	return (s_info.last_error == RES_OK);
}

/**
 * @brief Internal function which sets the error message to the display
 * based on the last_error member of the s_info internal structure.
 */
static void _set_display_error(void)
{
	s_info.error = true;
	switch (s_info.last_error) {
	case RES_NUMER_IS_TOO_LONG:
		display_set_str(STR_DISP_NUMBER_IS_TOO_LONG);
		break;
	case RES_INTERNAL_ERROR:
		display_set_str(STR_DISP_INTERNAL_ERROR);
		break;
	case RES_DIVISION_BY_ZERO:
		display_set_str(STR_DISP_DIVISION_BY_ZERO);
		break;
	case RES_UNDEFINED_VALUE:
		display_set_str(STR_DISP_UNDEFINED_VALUE);
		break;
	default:
		s_info.error = false;
		break;
	}
}

/**
 * @brief Internal function which unsets the the edit mode.
 */
void _unset_edit_mode(void)
{
	if (s_info.in_edit_mode) {
		s_info.in_edit_mode = false;
		s_info.num = display_get_number();
		s_info.num_is_new = true;
		/* This will remove extra chars from current display (for example: 2.300 -> 2.3; -0.0 -> 0) */
		display_set_number(s_info.num);
	}
}

/**
 * @brief Internal function which executes the chosen operation on provided value
 * and previously accumulated result.
 */
static bool _execute_operation(double rv)
{
	switch (s_info.op) {
	case KEY_ADD:
		s_info.acc += rv;
		break;
	case KEY_SUB:
		s_info.acc -= rv;
		break;
	case KEY_MUL:
		s_info.acc *= rv;
		break;
	case KEY_DIV:
		if (fabs(rv) <= CALC_ZERO_EPS)
			return ((fabs(s_info.acc) <= CALC_ZERO_EPS) ? RES_UNDEFINED_VALUE : RES_DIVISION_BY_ZERO);

		s_info.acc /= rv;
		break;
	default:
		s_info.last_error = RES_INTERNAL_ERROR;
		return false;
	}

	s_info.last_error = display_set_number(s_info.acc);
	if (s_info.last_error != RES_OK) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Failed to set display number: %f", s_info.acc);
		return false;
	}

	/* Get number back from display to get rounded value */
	s_info.num = display_get_number();
	s_info.acc = s_info.num;

	return true;
}

/**
 * @brief Internal function which performs negation of accumulated value.
 */
void _negate_value(void)
{
	if (s_info.num != 0.0) {
		s_info.num = -s_info.num;
		(void)display_set_number(s_info.num);
	}

	s_info.num_is_new = true;
}
