/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "display.h"
#include "$(appName).h"

#define MAX_MULTIPLIER 1.0e+11 /* 10 ^ (DISP_MAX_DIGITS - 1) */
#define NUMBER_ABS_LIMIT 1.0e+12 /* 10 ^ DISP_MAX_DIGITS */

static struct display_info {
	char str[DISP_STR_SIZE];
	int len;
	int dig_count;
	bool is_neg;
	bool has_dec_point;
	char dec_point_char;
	display_changed_cb change_cb;
} s_info = {
	.str = {0,},
	.len = 0,
	.dig_count = 0,
	.is_neg = false,
	.has_dec_point = false,
	.dec_point_char = '.',
	.change_cb = NULL,
};

static void _reset_props(void);
static void _set_display_string(const char *new_str);
static bool _is_empty(void);
static void _notify_change(void);
static long long _round(double value);

/**
 * @brief Registers the callback function to be invoked on display string change.
 * @param[in] cb The callback function pointer (NULL to unregister callback).
 */
void display_set_changed_cb(display_changed_cb cb)
{
	s_info.change_cb = cb;
}

/**
 * @brief Resets display to "0".
 */
void display_reset(void)
{
	_reset_props();
	_set_display_string("0");
}

/**
 * @brief Sets new decimal pointer character.
 * @param[in] new_dec_point_char New character to use as a decimal point.
 */
void display_set_dec_point_char(char new_dec_point_char)
{
	char *pos;

	if (new_dec_point_char != s_info.dec_point_char) {
		if (s_info.has_dec_point) {
			pos = strchr(s_info.str, s_info.dec_point_char);
			if (pos) {
				*pos = new_dec_point_char;
				_notify_change();
			}
		}

		s_info.dec_point_char = new_dec_point_char;
	}
}

/**
 * @brief Sets double value number to the display.
 * Number will be rounded according to the DISP_MAX_DIGITS value.
 * @param[in] value Double number to set to the display.
 * @return Result of the operation. If fail - object's state is not changed.
 */
result_type display_set_number(double value)
{
	char new_str[DISP_STR_SIZE] = {'\0'};
	bool is_neg = (value < 0.0);
	long long whole = (long long)(value);
	int whole_len = (whole != 0) ? (snprintf(new_str, DISP_STR_SIZE, "%lld", whole) - is_neg) : 1;
	int fract_len = DISP_MAX_DIGITS - whole_len;
	long long round_mant = 0;
	long long mult = 1;
	int i;
	int c;

	if (fabs(value) >= NUMBER_ABS_LIMIT) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Number is too long: %f", value);
		return RES_NUMER_IS_TOO_LONG;
	}

	if (whole == 0) {
		if (is_neg)
			round_mant = _round((value - 1.0) * MAX_MULTIPLIER);
		else
			round_mant = _round((value + 1.0) * MAX_MULTIPLIER);

		snprintf(new_str + 1, DISP_STR_SIZE - 1, "%lld", round_mant);
		 --new_str[is_neg + 1];
	} else if (fract_len >= 0) {
		for (i = 0; i < fract_len; i++)
			mult *= 10;

		round_mant = _round(value * (double)(mult));
		c = snprintf(new_str + 1, DISP_STR_SIZE - 1, "%lld", round_mant);
		if (c > DISP_MAX_DIGITS + is_neg) {
			++whole_len;
			--fract_len;
			new_str[c] = '\0';
		}
	}

	if (fract_len < 0) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Number is too long: %f", value);
		return RES_NUMER_IS_TOO_LONG;
	}

	s_info.len = DISP_MAX_DIGITS + is_neg + 1;
	s_info.dig_count = DISP_MAX_DIGITS;
	s_info.is_neg = is_neg;
	s_info.has_dec_point = true;

	memmove(new_str, new_str + 1, whole_len + is_neg);
	new_str[whole_len + is_neg] = s_info.dec_point_char;

	while (new_str[s_info.len - 1] == '0') {
		--s_info.len;
		--s_info.dig_count;
		new_str[s_info.len] = '\0';
	}
	if (s_info.dec_point_char == new_str[s_info.len - 1]) {
		--s_info.len;
		new_str[s_info.len] = '\0';
		s_info.has_dec_point = false;
	}
	if (is_neg && (s_info.len == 2) && (new_str[1] == '0')) {
		new_str[0] = '0';
		new_str[1] = '\0';
		s_info.len = 1;
		s_info.is_neg = false;
	}

	_set_display_string(new_str);

	return RES_OK;
}

/**
 * @brief Gets number from display as double value.
 * @return Double number of the string on display or 0.0 if it is not a number.
 */
double display_get_number(void)
{
	return strtod(s_info.str, NULL);
}

/**
 * @brief Processes number key category to enter new characters to display.
 * @param[in] key key Value to process.
 * @param[in] reset True will indicate that display should be reset before key process.
 * @return Result of the operation. If fail - object's state is not changed.
 */
result_type display_enter_key(key_type key, bool reset)
{
	char new_str[DISP_STR_SIZE] = {'\0'};
	char dig_char;

	if ((KEY_DEC != key) && ((key < KEY_0) || (key > KEY_9))) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Invalid key: %d", key);
		return RES_ILLEGAL_ARGUMENT;
	}

	if (reset) {
		_reset_props();
		strcpy(new_str, "0");
	} else {
		if (s_info.dig_count == DISP_MAX_DIGITS) {
			dlog_print(DLOG_ERROR, LOG_TAG, "Display is full");
			return RES_MAX_DIGITS_REACHED;
		}

		strcpy(new_str, s_info.str);
	}

	if (KEY_DEC != key) {
		dig_char = '0' + (key - KEY_0);
		if ((s_info.len == 1) && (new_str[0] == '0')) {
			new_str[0] = dig_char;
		} else {
			new_str[s_info.len] = dig_char;
			++s_info.len;
			new_str[s_info.len] = '\0';
			++s_info.dig_count;
		}
	} else {
		if (s_info.has_dec_point) {
			dlog_print(DLOG_ERROR, LOG_TAG, "Double decimal point entered");
			return RES_DOUBLE_DOT;
		}

		new_str[s_info.len] = s_info.dec_point_char;
		++s_info.len;
		new_str[s_info.len] = '\0';
		s_info.has_dec_point = true;
	}

	_set_display_string(new_str);

	return RES_OK;
}

/**
 * @brief Erases last entered character.
 * @return Result of the operation. If fail - object's state is not changed.
 */
result_type display_erase(void)
{
	if (_is_empty()) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Display is empty");
		return RES_DISPLAY_IS_EMPTY;
	}

	if (s_info.dig_count == 1 && (!s_info.has_dec_point || s_info.str[s_info.is_neg] == '0')) {
		display_reset();
		return RES_OK;
	}

	--s_info.len;

	if (s_info.has_dec_point && (s_info.str[s_info.len] == s_info.dec_point_char))
		s_info.has_dec_point = false;
	else
		--s_info.dig_count;

	s_info.str[s_info.len] = '\0';
	_notify_change();

	return RES_OK;
}

/**
 * @brief Negates number on display.
 */
void display_negate(void)
{
	if (!_is_empty()) {
		if (s_info.is_neg) {
			memmove(s_info.str, s_info.str + 1, s_info.len);
			s_info.len--;
			s_info.is_neg = false;
		} else {
			++s_info.len;
			memmove(s_info.str + 1, s_info.str, s_info.len);
			s_info.str[0] = '-';
			s_info.is_neg = true;
		}

		_notify_change();
	}
}

/**
 * @brief Sets string message to display.
 * After this function display should be reset to work correctly.
 * @param[in] new_str String to set to the display.
 */
void display_set_str(const char *new_str)
{
	if (!new_str) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Invalid input argument.");
		return;
	}

	_set_display_string(new_str);
}

/**
 * @brief Gets the string currently displayed.
 * @return Currently displayed string.
 */
const char *display_get_str(void)
{
	return s_info.str;
}

/**
 * @brief Internal function which resets the internal data structure to the initial state.
 */
static void _reset_props(void)
{
	s_info.len = 1;
	s_info.dig_count = 1;
	s_info.is_neg = false;
	s_info.has_dec_point = false;
}

/**
 * @brief Internal function which sets the given string to the display.
 * @param[in] new_str The string to be set to the display.
 */
static void _set_display_string(const char *new_str)
{
	if (strcmp(new_str, s_info.str) != 0) {
		strcpy(s_info.str, new_str);
		_notify_change();
	}
}

/**
 * @brief Internal function which checks whether the display string is empty or not.
 * @return This function returns 'true' if the display string is empty,
 * otherwise 'false' is returned.
 */
static bool _is_empty(void)
{
	return (s_info.len == 1 && s_info.str[0] == '0');
}

/**
 * @brief Internal function which invokes the 'changed' callback.
 */
static void _notify_change(void)
{
	if (s_info.change_cb)
		s_info.change_cb();
}

/**
 * @brief Internal function which performs round operation to the nearest integer value.
 * @return The nearest integer value is returned.
 */
static long long _round(double value)
{
	long long whole = (long long)(value);
	double fract = value - (double)(whole);

	if (fract >= 0.5)
		return whole + 1;
	else if (fract < -0.5)
		return whole - 1;

	return whole;
}
